#include "license_lua_userdata.h"
#include "xalloc.h"
#include "string_list.h"
#include "metric_list.h"
#include "license_logging.h"
#include "license_env.h"
#include "license_action.h"
#include <lua.h>
#include <lauxlib.h>

/* forward declarations*/
static int op5lic_query_env_udata_get_environment(lua_State *L);
static int op5lic_query_response_udata_gc(lua_State *L);
static int op5lic_query_response_udata_add_message(lua_State *L);
static int op5lic_query_response_udata_add_metric(lua_State *L);
static int op5lic_libfunc_subset(lua_State *L);

/* Userdata: env*/
struct op5lic_query_env_udata_t * op5lic_query_env_udata_new(lua_State *L,
		const struct op5lic_env_list_t * environment) {
	struct op5lic_query_env_udata_t *qr = (struct op5lic_query_env_udata_t *)
		lua_newuserdata(L, sizeof(struct op5lic_query_env_udata_t));
	luaL_getmetatable(L, "op5lic.environment");
	lua_setmetatable(L, -2);
	qr->environment = environment;
	return qr;
}

static const struct luaL_reg op5lic_env_udata_methods [] = {
	{"get", op5lic_query_env_udata_get_environment},
	{NULL, NULL} /* sentinel */
};


int op5lic_query_luaregister_userdata_env(lua_State *L) {
	luaL_newmetatable(L, "op5lic.environment");
	lua_pushstring(L, "__index");
	lua_pushvalue(L, -2);
	lua_settable(L, -3);

	luaL_register(L, NULL, op5lic_env_udata_methods);
	return 1;
}



static struct op5lic_query_env_udata_t * check_op5lic_query_env_udata(
		lua_State *L)  {
	void *udata = luaL_checkudata(L, 1, "op5lic.environment");
	luaL_argcheck(L, udata != NULL, 1, "`environment' expected");
	return (struct op5lic_query_env_udata_t *) udata;
}

static int op5lic_query_env_udata_get_environment(lua_State *L) {
	struct op5lic_env_node_t * node = NULL;
	struct op5lic_query_env_udata_t *qr = check_op5lic_query_env_udata(L);
	const char *env_path = luaL_checkstring(L, 2);
	luaL_argcheck(L, env_path != NULL, 2, "`string' expected");
	node = op5lic_env_list_find(qr->environment, env_path);

	if (!node) {
		log_info("Could not find `%s' in current environment", env_path);
		if (lua_type(L, 3) == LUA_TNONE) {
			lua_pushnil(L);
		}
	}
	else {
		switch (node->type) {
			case ENV_TYPE_INTEGER:
				lua_pushinteger(L, node->value.val_int);
				break;
			case ENV_TYPE_ENV_LIST:
				lua_pushnil(L); /* makes sense? */
				break;
			case ENV_TYPE_STRING:
				lua_pushstring(L, node->value.val_string);
				break;
		}
	}
	return 1;
}


/* Userdata: Response */

struct op5lic_query_response_udata_t * op5lic_query_response_udata_new(lua_State *L) {
	struct op5lic_query_response_udata_t *qr = (struct op5lic_query_response_udata_t *)
		lua_newuserdata(L, sizeof(struct op5lic_query_response_udata_t));
	luaL_getmetatable(L, "op5lic.response");
	lua_setmetatable(L, -2);
	qr->message_list = string_list_new();
	qr->metric_list = metric_list_new();
	return qr;
}

static const struct luaL_reg op5lic_response_udata_methods [] = {
	{"add_message", op5lic_query_response_udata_add_message},
	{"add_metric", op5lic_query_response_udata_add_metric},
	{NULL, NULL} /* sentinel */
};

int op5lic_query_luaregister_userdata_response(lua_State *L) {
	luaL_newmetatable(L, "op5lic.response");

	lua_pushstring(L, "__index");
	lua_pushvalue(L, -2);
	lua_settable(L, -3);

	lua_pushstring(L, "__gc");
	lua_pushcfunction(L, op5lic_query_response_udata_gc);
	lua_settable(L, -3);

	luaL_register(L, NULL, op5lic_response_udata_methods);
	return 1;
}

static struct op5lic_query_response_udata_t * check_op5lic_query_response_udata(
		lua_State *L)  {
	void *udata = luaL_checkudata(L, 1, "op5lic.response");
	luaL_argcheck(L, udata != NULL, 1, "`response' expected");
	return (struct op5lic_query_response_udata_t *) udata;
}

/*
 * Adds a message to the result userdata
 * Lua-Usage: response_obj:add_message(message)
 */
static int op5lic_query_response_udata_add_message(lua_State *L) {
	struct op5lic_query_response_udata_t *qr = check_op5lic_query_response_udata(L);
	const char * message = luaL_checkstring(L, 2); /* the message*/
	luaL_argcheck(L, message != NULL, 2, "`string' expected");

	string_list_add(qr->message_list, message);
	log_debug("Added message `%s' to message list", message);
	return 0;
}

/*
 * Adds a metric to the result userdata
 * Lua-Usage: response_obj:add_metric(id, value, min, max)
 */
static int op5lic_query_response_udata_add_metric(lua_State *L) {
	struct op5lic_query_response_udata_t *qr = check_op5lic_query_response_udata(L);
	int value, min, max;
	const char * id = luaL_checkstring(L, 2);
	value = luaL_checkint(L, 3);
	min = luaL_checkint(L, 4);
	max = luaL_checkint(L, 5);
	if (metric_list_add(qr->metric_list, id, value, min, max) == NULL) {
		log_err("Failed to add metric `%s' to metric list", id);
	}
	return 0;
}


static int op5lic_query_response_udata_gc(lua_State *L) {
	struct op5lic_query_response_udata_t *qr = check_op5lic_query_response_udata(L);
	string_list_destroy(qr->message_list);
	metric_list_destroy(qr->metric_list);
	return 0;
}


static const struct luaL_reg op5lic_libfunctions [] = {
	{"subset", op5lic_libfunc_subset},
	{NULL, NULL} /*sentinel*/

};

int op5lic_query_luaregister_libfunctions(lua_State *L) {
	luaL_register(L, "op5lic", op5lic_libfunctions);
	return 1;
}

static int op5lic_libfunc_subset(lua_State *L) {
	const char *a = luaL_checkstring(L, 1);
	const char *b = luaL_checkstring(L, 2);
	lua_pushboolean(L, op5lic_license_action_subsetof(a, b));
	return 1;
}
